/*
 * Functions for Form2mail that generates an image from a security key,
 * to distinguish between humans and robots.
 *
 * Based by obfuscation.c (C) 2005 Muthiah Annamalai <gnumuthu@users.sf.net>
 * National Institute of Technology,
 * Tiruchirapalli, India
 *
 * Copyright (C) 2007 Vladimir Oleynik <dzo@simtreas.ru>
 *  - rewrote for make captcha from Obfuscation.c by Muthiah
 *  - many clearing
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <gd.h>
#include <unistd.h>
#include <errno.h>

#include "loadcfg.h"

static struct CFG form_cfg[] = {
	/* default */
	{ "database_name", "captcha.dbc"},              /* 0 */
	{ "GDFONTPATH", "/usr/share/fonts/truetype"},   /* 1 */
	{ "default_font", "gothicb.ttf" },              /* 2 */
	{ NULL, NULL}
};


#define DATABASE_NAME           form_cfg[0].value
#define GDFONTPATH_CFG          form_cfg[1]
#define DEFAULT_FONT            form_cfg[2].value

#define CHAR_SIZE 50
#define MAX_CHARS 7

#define BLACK 0x000000
#define WHITE 0xffffff
#define BABYBLUE 0x6699CC

#define FG BLACK
#define BG 0xABCDEF
#define BORDER BABYBLUE


static int rand_rgb(void)
{
  static int valid_rgb[6] = { 0x00, 0x33, 0x66, 0x99, 0xCC, 0xFF };
  int rgb[3];
  int i;

  for(i = 0; i < 3; i++)
	rgb[i] = valid_rgb[rand() % 6];

  return rgb[0] + (rgb[1] << 8) + (rgb[2] << 16);
}


static void
randImage (const char *text)
{
  gdImagePtr ptr;
  int iter;
  char str[2];
  int brect[8];
  double angle;
  int prevX = 25, prevY = 25;
  int max_chars = strlen(text);
  FILE *out;

  memset(brect, 0, sizeof(brect));
  if(max_chars < 2)
	max_chars = 2;

  str[1] = '\0';

  ptr = gdImageCreateTrueColor (CHAR_SIZE * (max_chars - 1), CHAR_SIZE);
  if(ptr == NULL) {
	perror("gdImageCreateTrueColor");
	exit(1);
  }


  /* Fill Background */
  gdImageFill (ptr, 0, 0, BG);

  for (iter = 0; iter < max_chars; iter++) {
	int temp;
	int oldX = rand () % 15 + 10;

	str[0] = text[iter];

	brect[0] = 0;
	brect[1] = 0;
	brect[2] = 25;
	brect[3] = 25;

	oldX += iter * (4.0 / 5.0) * CHAR_SIZE;
	temp = rand() & 0x07;

	switch (temp) {
	case 0:
	  angle = -55.0;
	  break;
	case 1:
	  angle = -35.0;
	  break;
	case 2:
	  angle = -25.0;
	  break;
	case 3:
	  angle = -15.0;
	  break;
	case 4:
	  angle = 20.0;
	  break;
	case 5:
	  angle = 40.0;
	  break;
	case 6:
	  angle = 60.0;
	  break;
	default:
	  angle = 50.0;
	}

	angle = angle / 360.0;
	angle = angle * (22.000 / 7.000);

	gdImageStringFT (ptr, brect, FG, DEFAULT_FONT, 20.0, angle, oldX,
					 10 + CHAR_SIZE / 2, str);

	gdImageSetThickness (ptr, rand() % 2 + 1);
	gdImageLine(ptr, prevX, prevY, brect[0], (brect[1] + brect[3]) / 3.0,
				rand_rgb());
	prevX = brect[0];
	prevY = (brect[3] + brect[1]) / 3.0;

	if(DEBUG)
		fprintf (stderr, "%g (%d,%d)->(%d, %d)\n", angle,
				brect[0], brect[1], brect[2], brect[3]);

	/* Break the character */
	gdImageSetThickness (ptr, 2);
	gdImageLine (ptr, brect[0], 0, brect[2], CHAR_SIZE, rand_rgb());

	switch (str[0]) {
	case 'i':
	case 'I':
	case 'j':
	case 'J':
	  /* case '1': */
	  /* Dont Draw a second line for these characters */
	  break;
	default:
	  gdImageLine (ptr, brect[2], 0, brect[0], CHAR_SIZE, rand_rgb());
	  gdImageSetThickness (ptr, 1);
	}
  }

  /* Draw Border */
  gdImageSetThickness (ptr, 5);
  gdImageRectangle (ptr, 0, 0, CHAR_SIZE * (max_chars - 1) - 1, CHAR_SIZE - 1,
					BORDER);

  /* Draw a buffer-White Border */
  gdImageSetThickness (ptr, 2);
  gdImageRectangle (ptr, 5, 5, CHAR_SIZE * (max_chars - 1) - 5, CHAR_SIZE - 5,
			WHITE);

  if(DEBUG) {
	out = fopen("captcha_debug.gif", "wb");
	if(out == NULL) {
		perror("captcha_debug.gif");
		exit(5);
	}
	fprintf(stderr, "%s saved\n", "captcha_debug.gif");
  } else {
	printf("Content-type: image/gif\n\n");
	out = stdout;
  }
  gdImageGif (ptr, out);
  gdImageDestroy(ptr);
}

/*
  :

key=   int_32

time_t
int_32 -   (24 )
*/


static int get_db(char text[MAX_CHARS + 1], uint32_t db_key)
{
	int fd;
	time_t tc;
	int32_t k, code;
	const size_t sz = sizeof(db_key)+sizeof(tc)+sizeof(code);
	char content[sz];
	ssize_t r;

	if((fd = open_db_reader(DATABASE_NAME)) < 0) {
		printf("Content-type: text/plain\n\n");
		printf("Database problem\n");
		return 3;
	}
	while(1) {
		r = read(fd, content, sz);
		if(r == 0) {
			printf("Content-type: text/plain\n\n");
			printf("Expired?\n");
			return 4;
		}
		if(r < 0 && errno == EINTR)
			continue;
		if(r < sz) {
			printf("Content-type: text/plain\n\n");
			printf("Database problem, corrupted?\n");
			return 1;
		}
		memcpy(&k, content, sizeof(k));
		if(k == db_key)
			break;
	}
	close(fd);
	memcpy(&tc, content + sizeof(k), sizeof(tc));
	memcpy(&code, content + sizeof(k) + sizeof(tc), sizeof(code));
	sprintf(text, "%d", code);
	srand(tc + getpid());
	return 0;
}

int
main (int argc, char *argv[])
{
  char *q, *e;
  char text[MAX_CHARS + 1];
  int i;
  long db_key;


  if(LoadConfig(form_cfg) < 0) {
	printf("Content-type: text/plain\n\n");
	printf("LoadConfig problem\n");
	return 14;
  }
  e = malloc(strlen(GDFONTPATH_CFG.name) + strlen(GDFONTPATH_CFG.value) + 2);
  if(e == NULL) {
	printf("Content-type: text/plain\n\n");
	printf("memory exchausted\n");
	return 2;
  }
  sprintf(e, "%s=%s", GDFONTPATH_CFG.name, GDFONTPATH_CFG.value);
  /* putenv("GDFONTPATH=/usr/share/fonts/truetype"); */
  putenv(e);

  q = getenv("QUERY_STRING");
  if(q == NULL) {
	printf("Content-type: text/plain\n\n");
	printf("Empty query\n");
	return 10;
  }
  db_key = strtol(q, &e, 10);
  if(db_key <= 0 || *e != '\0' ||
		(long long)db_key > 0xffffffffll) {
	printf("Content-type: text/plain\n\n");
	printf("Invalid query\n");
	return 11;
  }

  i = get_db(text, db_key);
  if(i == 0)
	randImage (text);
  else
	return i;
  return 0;
}
