/*
 * Functions for Form2mail that
 * - generate crypt key, store to database
 * - read form html file, replace (added) db key for compare with secret key
 *
 *
 * Copyright (C) 2007 Vladimir Oleynik <dzo@simtreas.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <stdint.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <crypt.h>
#include <errno.h>

#include "loadcfg.h"


static struct CFG form_cfg[] = {
	/* default */
	{ "Database_Name", "captcha.dbc"},      /* 0 */
	{ "added_secret", "Maci v1.0"},         /* 1 */
	{ "find_word", "<HERE>"},               /* 2 */
	{ "form_with_captha", "form.html" },    /* 3 */
	{ NULL, NULL }
};

#define DATABASE_NAME           form_cfg[0].value
#define ADDED_SECRET            form_cfg[1].value
#define FIND_WORD               form_cfg[2].value
#define FORM_WITH_CAPTHA        form_cfg[3].value

/*
  :

key=   int_32

time_t
int_32 -   (24 )
*/

static char secret[32]; // strlen(ADDED_SECRET)+10 min
static time_t curr_time;

static int find_expires_and_create_next(void)
{
	time_t tc;
	int fd;
	int32_t code;
	uint32_t k, kmax = 0;
	const size_t sz = sizeof(k)+sizeof(tc)+sizeof(code);
	char content[sz];
	char *encoded;
	ssize_t r;
	off_t first_expired = 0;

	time (&curr_time);
	if((fd = open_db_writer (DATABASE_NAME)) < 0)
		return 3;
	while(1) {
		r = read(fd, content, sz);
		if(r == 0)
			break;
		if(r < 0 && errno == EINTR)
			continue;
		if(r < sz) {
			// hmm, corrupted?
			return 1;
		}
		memcpy(&k, content, sizeof(k));
		if(kmax < k)
			kmax = k;
		if(first_expired == 0) {
			memcpy(&tc, content + sizeof(k), sizeof(tc));
			if((tc+60) < curr_time)
				first_expired = lseek(fd, (off_t)0, SEEK_CUR);
		}
	}
	kmax++;
	if(first_expired)
		lseek(fd, first_expired - sz, SEEK_SET);
	// make new code
	sprintf(secret, "%x%s", kmax, ADDED_SECRET);
	encoded = crypt(secret, "$1$salt$");   // $1$salt$<MD5>
	code = ((int)encoded[10]) + (((int)encoded[9]) << 8) +
			(((int)encoded[8]) << 16);
	sprintf(secret, "%d", kmax);
	// store new record
	memcpy(content, &kmax, sizeof(kmax));
	memcpy(content+sizeof(kmax), &curr_time, sizeof(curr_time));
	memcpy(content+sizeof(kmax)+sizeof(curr_time), &code, sizeof(code));
	write(fd, content, sz);
	close (fd);
	return 0;
}

static int
loop_work (FILE *in, int lfb, char *find_buf)
{
  int c;
  int i = 0;

  for (;;) {
	/* fast first char compare */
	c = getc (in);
	if (c == EOF) {
	  if (ferror (in))
		return -1;
	  return 0;
	}
	if (c != FIND_WORD[0]) {
	  if (putchar (c) == EOF)
		return -1;
	  continue;
	}
	*find_buf = c;
	i = 1;

	/* fast next chars compare */
	while (i < lfb) {
	  c = getc (in);
	  if (c == EOF) {
	      if (ferror (in))
		return -1;
	      c = 0;
	      while (i) {
		if (putchar (find_buf[c++]) == EOF)
		  return -1;
		i--;
	      }
	      return 0;
	  }
	  find_buf[i] = c;
	  if(FIND_WORD[i++] != c) {
loop2:
		if (putchar (*find_buf) == EOF)
			return -1;
		for (c = 1; c < i; c++)
			find_buf[c - 1] = find_buf[c];
		if (--i == 0)
			break;  /* to fast first char compare */
		if (strncmp(find_buf, FIND_WORD, i) == 0)
			continue;       /* to fast next chars compare */
		goto loop2;
	  }
	}
	if(i == lfb) {
		// all compared
		if(fputs(secret, stdout) == EOF)
			return -1;
	}
  }
  return i == lfb;
}

int main(int argc, char **argv)
{
	int r;

	if(LoadConfig(form_cfg) < 0) {
		printf("Content-type: text/plain\n\n");
		printf("LoadConfig problem\n");
		return 14;
	}
	r = find_expires_and_create_next();
	if(r) {
		printf("Content-type: text/plain\n\n");
		printf("Database %s problem %d\n", DATABASE_NAME, r);
		return 4;
	}
	{
		FILE *in;
		int lfb;
		char *find_buf;
		int ret;
		char ctime_buf[32];
		char expired_buf[32];

		in = fopen(FORM_WITH_CAPTHA, "r");
		if(in == NULL) {
			printf("Content-type: text/plain\n\n");
			printf("Form not found\n");
			return 3;
		}
		lfb = strlen(FIND_WORD);
		find_buf = malloc(lfb);

		if(find_buf == NULL) {
			printf("Content-type: text/plain\n\n");
			printf("memory exchausted\n");
			return 2;
		}
		setenv("TZ", "", 1);
		tzset();
		strcpy(ctime_buf, ctime(&curr_time));
		/* Mon Jan 14 12:06:41 2008 */
		ctime_buf[ 3] = '\0';
		ctime_buf[ 7] = '\0';
		ctime_buf[10] = '\0';
		ctime_buf[19] = '\0';
		ctime_buf[24] = '\0';
		/* Fri, 28 Jul 2006 15:30:00 UTC */
		sprintf(expired_buf, "%s, %s %s %s %s UTC",
			ctime_buf, ctime_buf+8, ctime_buf+4,
			ctime_buf+20, ctime_buf+11);
		printf("Expires: %s\n", expired_buf);
		printf("Content-type: text/html\n\n");
		// show_form
		ret = loop_work (in, lfb, find_buf);
		if(ret == 1)
			return 0;
		if(ret < 0)             // IO error
			return 5;
		return 6;               // FIND_WORD not found
	}
}
