/*
 * Functions for Form2mail that
 * - parse input form
 * - compare captcha
 * - call sendmail with head,body and attached (optional) file
 *
 * Copyright (C) 2007 Vladimir Oleynik <dzo@simtreas.ru>
 *  - based by Upload 2.6 (C) Jeroen C. Kessels <jeroen@kessels.com> 2000
 *  - many clearing, reduce memory (zilion kb stack usage)
 *
 *  Base64 encoder. Stolen from busybox and optimize by
 *  (C) 2003 Vladimir Oleynik <dzo@simtreas.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */


#define _GNU_SOURCE
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>

#include "loadcfg.h"

#define MAX_Context_field 1024

static struct CFG form_cfg[] = {
	/* default */
	{ "database_name", "captcha.dbc"},      /* 0 */
	{ "Tmp_for_files", "upload" },          /* 1 */
	{ "BadPage", "BadPage.html" },          /* 2 */
	{ "OkPage", "OkPage.html" },            /* 3 */
	{ "LogFile", "" },                      /* 4 */
	{ "FormHidden",    "num"  }, /*    ()   5 */
	{ "FormLastName",  "lname"}, /*                 6 */
	{ "FormFirstName", "fname"}, /*                     7 */
	{ "FormSecName", "sname"  }, /*                8 */
	{ "FormAddress", "address"}, /*           9 */
	{ "FormPhone", "phone"    }, /*                10 */
	{ "FormFrom", "email"     }, /* Email                 11 */
	{ "FormSubj", "subj"      }, /*          12 */
	{ "FormBody", "body"      }, /*              13 */
	{ "FormCaptcha", "code"   }, /*    14 */
	{ "NonEmpty",    "FormAddress+FormPhone" },   /*      15 */
	{ "PageIsNotComplete", "PageIsNotComplete.html" }, /* 16 */
	{ "ExpiredPage", "ExpiredPage.html" },             /* 17 */
	{ "charset", "koi8-r" },                           /* 18 */
	{ "X-Mailer", "Form2Mail by <dzo@simtreas.ru> (U; Linux 2.4.33 i686)" },
	{ "To", "dzo@simtreas.ru" },                       /* 20 */
	{ "From", "\" \" <nobody@simtreas.ru>" },/* 21 */
	{ "Subject", "   " },    /* 22 */
	{ "Sendmail", "/usr/sbin/sendmail" },              /* 23 */
	{ NULL, NULL }
};

#define DATABASE_NAME           form_cfg[0].value
#define NonEmptyCfg             15
#define PageIsNotComplete       form_cfg[16].value
#define ExpiredPage             form_cfg[17].value
#define Tmp_for_files           form_cfg[1].value
#define BadPage                 form_cfg[2].value
#define OkPage                  form_cfg[3].value
#define LogFile                 form_cfg[4].value
#define charset                 form_cfg[18].value
#define X_Mailer                form_cfg[19].value
#define To                      form_cfg[20].value
#define From                    form_cfg[21].value
#define Subject                 form_cfg[22].value
#define Sendmail                form_cfg[23].value

struct FORM {
	int cfg_num;
	char *stored;
	int flg_must_non_empty;
};

static struct FORM Form[] = {
	{  5, NULL, 1 },        /* 0 */
	{  6, NULL, 1 },        /* 1 */
	{  7, NULL, 1 },        /* 2 */
	{  8, NULL, 0 },        /* 3 */
	{  9, NULL, 1 },        /* 4 */
	{ 10, NULL, 1 },        /* 5 */
	{ 11, NULL, 0 },        /* 6 */
	{ 12, NULL, 0 },        /* 7 */
	{ 13, NULL, 1 },        /* 8 */
	{ 14, NULL, 1 },        /* 9 */
	{  0, NULL, 0 }         /* 10 */
};
#define Form_num 0
#define Form_try 9

#define Start_Form_send_to_email_body 1
#define End_Form_send_to_email_body   8


/* base 64 buffers size */
#define SRC_BUF_SIZE 54                 /* 4/3 set output line 72 size */
#if (SRC_BUF_SIZE/3)*3 != SRC_BUF_SIZE
# error SRC_BUF_SIZE *MUST* be a multiple of 3
#endif
#define DST_BUF_SIZE  4 * (SRC_BUF_SIZE / 3)
static char src_buf[SRC_BUF_SIZE];
static char dst_buf[DST_BUF_SIZE + 1];       /* + '\n' */

static char FieldName[BUFSIZ];  /* Name of the variable from the form. */
static char UpFileName[BUFSIZ];
static char Attached[BUFSIZ];
static char LastFileName[BUFSIZ];

static char *Boundary;  /* Character string that delimits segments. */
static size_t strlen_Boundary;

static void atexit_fun(void)
{
	if(!DEBUG)
		unlink(LastFileName);
}

/* Show the BadPage, with macro MESSAGE. */
static void ShowBadPage (const char *fmt, ...) __attribute__((noreturn));

static void
ShowBadPage (const char *fmt, ...)
{
  FILE *Fin;
  char Line[BUFSIZ];
  char *p1;
  va_list p;

  if (DEBUG == 0)
	printf ("Content-type: text/html\n\n");

  va_start(p, fmt);
  Fin = fopen (BadPage, "r");
  if (Fin == NULL) {
	printf ("<html>\n<body>\n");
	if (BadPage[0] != '\0') {
	  printf ("Error: could not open the BadPage: %s<p>\n", BadPage);
	  printf ("The original error message was:<br>\n");
	}
	vprintf(fmt, p);
	printf ("\n</body>\n</html>\n");
	exit (1);
  }

  while (fgets (Line, BUFSIZ, Fin) != NULL) {
	p1 = Line;
	while (*p1 != '\0') {
	  if (strncasecmp (p1, "<insert message>", 15) == 0) {
		vprintf(fmt, p);
		p1 += 16;
	  } else {
		putchar(*p1);
		p1++;
	  }
	}
  }

  exit (1);
}

/* Write a line to the logging file. */
static void
WriteLogLine (const char *fmt, ...)
{
  FILE *Fout;
  char s1[32];
  time_t Now;
  va_list p;

  if (LogFile[0] == '\0')
	return;

  Fout = fopen (LogFile, "a");
  if (Fout == NULL)
	ShowBadPage ("I could not open the logfile: %s", LogFile);

  Now = time (NULL);
  strncpy (s1, ctime (&Now), sizeof(s1));
  s1[24] = '\0';
  fprintf(Fout, "%s ", s1);
  va_start(p, fmt);
  vfprintf(Fout, fmt, p);
  putc('\n', Fout);
  fclose (Fout);
}


/* Show the OkPage, with statistics about the file. */
static void ShowOkPage (const char *Page) __attribute__((noreturn));
static void
ShowOkPage (const char *Page)
{
  FILE *Fin;
  char Line[BUFSIZ];
  char *p1;

  Fin = fopen (Page, "r");
  if (Fin == NULL)
	ShowBadPage ("I could not open the Page '%s'", Page);

  if (DEBUG == 0)
	printf ("Content-type: text/html\n\n");

  while (fgets (Line, BUFSIZ, Fin) != NULL) {
	p1 = Line;
	do {
		p1 = strcasestr(Line, "<insert lastfilename>");
		if(p1) {
			*p1 = '\0';
			printf ("%s%s", Line, LastFileName);
			strcpy(Line, p1 + 21);
		}
	} while(p1);
	if(Line[0] != '\0')
		printf ("%s", Line);
  }

  exit (0);
}

static void StoreFormField(const char *name, const char *value)
{
	struct FORM *f;

	for(f = Form; f->cfg_num != 0; f++) {
		if(strcasecmp(name, form_cfg[f->cfg_num].value) == 0) {
			f->stored = strdup(value);
			break;
		}
	}
}

/* Conversion table.  for base 64 */
static const char tbl_base64[] = {
  'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
  'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
  'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
  'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
  'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
  'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
  'w', 'x', 'y', 'z', '0', '1', '2', '3',
  '4', '5', '6', '7', '8', '9', '+', '/',
};


static int
b64encode (const char *from, const char *store, const int length)
{
  int i;
  unsigned char *p = (unsigned char *) store;
  const unsigned char *s = (const unsigned char *) from;

  /* Transform the 3x8 bits to 4x6 bits, as required by base64.  */
  for (i = 0; i < length; i += 3) {
	*p++ = tbl_base64[s[0] >> 2];
	*p++ = tbl_base64[((s[0] & 3) << 4) + (s[1] >> 4)];
	*p++ = tbl_base64[((s[1] & 0xf) << 2) + (s[2] >> 6)];
	*p++ = tbl_base64[s[2] & 0x3f];
	s += 3;
  }
  /* Pad the result if necessary...  */
  if (i == length + 1) {
	*(p - 1) = '=';
  } else if (i == length + 2) {
	*(p - 1) = *(p - 2) = '=';
  }
  /* ...and CR-terminate it.  */
  *p = '\n';

  return i / 3 * 4 + 1;
}


/* Accept a single segment from the incoming mime stream. Each field in the
   form will generate a mime segment. Return a pointer to the beginning of
   the Boundary, or NULL if the stream is exhausted. */
static char *
AcceptSegment (char *Input,             /* Pointer into the incoming stream. */
		long *InputLength)      /* Bytes left in the incoming stream. */
{
  char *ContentStart;                   /* Pointer to the file content. */
  char *ContentEnd;
  char *ContentAsString;
  long ContentLength;                   /* Bytecount of the content. */
  FILE *Fout;
  char *p1;
  int i;
  long il = *InputLength;

  *UpFileName = '\0';
  /* The input stream should begin with a Boundary line. Error-exit if not
     found. */
  if (strncmp (Input, Boundary, strlen_Boundary) != 0) {
	WriteLogLine ("Missing boundary '%s' in input.", Boundary);
	ShowBadPage ("Missing boundary '%s' in input.", Boundary);
  }

  /* Skip the Boundary line. */
  Input += strlen_Boundary;
  il -= strlen_Boundary;
  /* no more segments? */
  if (il == 0 || strncmp (Input, "--", 2) == 0)
	return NULL;

  if (strncmp (Input, "\r\n", 2) != 0) {
	p1 = "Missing CRLF in input.";
	WriteLogLine (p1);
	ShowBadPage (p1);
  }
  Input += 2;
  il -= 2;
  /* no more segments? */
  if (il == 0 || strncmp (Input, "--", 2) == 0)
	return NULL;

  /* The first line of a segment must be a "Content-Disposition" line. It
     contains the fieldname, and optionally the original filename. Error-exit
     if the line is not recognised. */
  p1 = "Content-Disposition: form-data; name=\"";
  i = strlen(p1);
  if(strncasecmp (Input, p1, i) != 0) {
	WriteLogLine ("\"Content-Disposition\" need");
	ShowBadPage ("\"Content-Disposition\" need");
  }
  Input += i;
  il -= i;

  p1 = Input;
  while(il > 0 && *p1 != '"') {
	  p1++;
	  il--;
  }
  *p1 = '\0';
  /* safe strcpy */
  strncpy(FieldName, Input, sizeof(FieldName) - 1);
  FieldName[sizeof(FieldName) - 1] = '\0';
  if(*FieldName == '\0' || il == 0) {
	  WriteLogLine ("strange FieldName format.");
	  ShowBadPage ("strange FieldName format");
  }
  il--;
  Input = p1 + 1;

  p1 = "; filename=\"";
  i = strlen(p1);
  if (strncasecmp (Input, p1, i) == 0) {
	Input += i;
	p1 = Input;
	il -= i;

	while(il > 0 && *p1 != '"') {
		p1++;
		il--;
	}
	*p1 = '\0';
	/* safe strcpy */
	strncpy(UpFileName, Input, sizeof(UpFileName) - 1);
	UpFileName[sizeof(UpFileName) - 1] = '\0';
	if(il == 0) {
		WriteLogLine ("strange FileName format.");
		ShowBadPage ("strange FileName format");
	}
	il--;
	Input = p1 + 1;
  }
  if (strncmp (Input, "\r\n", 2) != 0) {
	WriteLogLine ("Missing CRLF in input.");
	ShowBadPage ("Missing CRLF in input.");
  }
  Input += 2;
  il -= 2;
  if(strncmp (Input, "\r\n", 2) == 0) {
	Input += 2;
	il -= 2;
  } else {
	p1 = strstr(Input, "\r\n\r\n");
	if(p1 == NULL) {
		WriteLogLine ("Missing CRLF CRLF in input.");
		ShowBadPage ("Missing CRLF CRLF in input.");
	}
	Input = p1 + 4;
	il -= (p1 - Input) + 4;
  }

  /* The following data in the stream is binary. The Boundary string is the
     end of the data. There may be a CRLF just before the Boundary, which
     must be stripped. */
  ContentStart = Input;
  ContentLength = 0;
  while (il > strlen_Boundary && memcmp(Input, Boundary, strlen_Boundary) != 0) {
	Input++;
	il--;
	ContentLength++;
  }
  *InputLength = il;
  ContentEnd = Input - 1;
  if ((ContentLength > 0) && (*ContentEnd == '\n')) {
	ContentEnd--;
	ContentLength = ContentLength - 1;
  }
  if ((ContentLength > 0) && (*ContentEnd == '\r')) {
	ContentEnd--;
	ContentLength = ContentLength - 1;
  }
  ContentAsString = ContentStart;
  if(*UpFileName == '\0') {
    size_t sz = (ContentEnd-ContentStart) + 1;
    ContentAsString[sz < MAX_Context_field ? sz : MAX_Context_field - 1] = '\0';
  }

  /* Show debugging information. */
  if (DEBUG > 0) {
	printf (" <tr>\n");
	printf ("  <td valign=top rowspan=2>%s</td>\n", FieldName);
	printf ("  <td valign=top>");
	if (*UpFileName != '\0') {
	  printf ("%s", UpFileName);
	} else {
	  printf ("%s", ContentAsString);
	}
	printf ("</td>\n");
	printf ("  <td valign=top>%ld</td>\n", ContentLength);
	printf (" </tr>\n");
  }

  /* Do nothing if this is not a file, but some other kind of field. */
  if (*UpFileName == '\0') {
	if (DEBUG) {
	  printf ("<tr><td colspan=2>FieldName not recognised.</td></tr>\n");
	}
	if(ContentAsString[0])
		StoreFormField(FieldName, ContentAsString);
	return Input;
  }
  if(LastFileName[0]) {
	/* accept one file only */
	return Input;
  }
  /* Determine the filename to store the file. If the UpFileName field was
     defined then use it, otherwise use the name of the incoming file. The
     ROOT is always prepended. */
  p1 = strrchr (UpFileName, '/');
  if (p1 != NULL) {
	p1++;
	strcpy (UpFileName, p1);
  }
  while( (p1 = strchr (UpFileName, '"')) != NULL )
	*p1 = '_';
  strcpy(Attached, UpFileName);

  i = strlen(Tmp_for_files) + 22;       /* 64 bit max */
  if(i > sizeof(LastFileName)) {
	WriteLogLine ("Filename too long (%d)", i);
	ShowBadPage ("Filename too long.");
  }
  sprintf (LastFileName, "%s/%d", Tmp_for_files, getpid());

  /* Open the file for writing. */
  Fout = fopen (LastFileName, "w");
  if (Fout == NULL) {
	WriteLogLine ("The file could not be uploaded because "
			"write permission is denied.<br>Filename: %s", LastFileName);
	ShowBadPage ("The file could not be uploaded because "
			"write permission is denied.<br>Filename: %s", LastFileName);
  }
  /* remove uploaded attach after exit */
  atexit(atexit_fun);

  /* Write the file to disk into base64 code */
  while(ContentLength > 0) {
	size_t size;

	if (ContentLength < SRC_BUF_SIZE) {
		size = ContentLength;
		p1 = memcpy(src_buf, ContentStart, size);
		/* pad with 0s so we can just encode extra bits */
		memset (p1 + size, 0, 3 - (size % 3));
	} else {
		size = SRC_BUF_SIZE;
		p1 = ContentStart;
	}
	ContentStart += size;
	ContentLength -= size;
	/* Encode the buffer */
	size = b64encode (p1, dst_buf, size);
	if(fwrite (dst_buf, 1, size, Fout) != size) {
		WriteLogLine ("File uploaded couldnt finish writing: %m\n");
		ShowBadPage ("I/O error");
	}
  }

  fclose (Fout);

  /* Show debugging information. */
  if (DEBUG) {
	printf ("<tr>\n");
	printf ("  <td colspan=2>File written: %s</td>\n", LastFileName);
	printf ("  </tr>\n");
  }

  WriteLogLine ("File uploaded succesfully: %s (%s) (%lu bytes)",
		   LastFileName, Attached, ContentLength);
  return Input;
}

static void DecodeNonEmpty(void)
{
	char *p = strdup(form_cfg[NonEmptyCfg].value);
	char *pp;
	int i;

	if(p == NULL) {
		WriteLogLine ("Memory exchausted.");
		ShowBadPage ("Memory exchausted.");
	}
	while( (pp = strchr(p, '+')) != NULL) {
		*pp = '\0';
		for(i = 0; form_cfg[i].name; i++) {
			if(strcasecmp(p, form_cfg[i].name) == 0) {
				int m;
				for(m = 0; Form[m].cfg_num != 0; m++)
					if(Form[m].cfg_num == i) {
						Form[m].flg_must_non_empty = 1;
						break;
					}
			break;
			}
		}
		p = pp + 1;
	}

	for(i = 0; Form[i].cfg_num != 0; i++) {
		if(Form[i].flg_must_non_empty && Form[i].stored == NULL) {
			if(DEBUG) {
				printf("%s is null\n", form_cfg[Form[i].cfg_num].name);
			}
			ShowOkPage(PageIsNotComplete);
		}
	}
}

static int get_db(uint32_t db_key, uint32_t try_key)
{
	int fd;
	time_t tc;
	int32_t k, code;
	const size_t sz = sizeof(db_key)+sizeof(tc)+sizeof(code);
	char content[sz];
	ssize_t r;

	if((fd = open_db_writer(DATABASE_NAME)) < 0) {
		WriteLogLine ("Database problem: %m");
		ShowBadPage ("Database problem");
	}
	while(1) {
		r = read(fd, content, sz);
		if(r == 0) {
			WriteLogLine ("Code %d expired?", db_key);
			if(!DEBUG)
				ShowOkPage (ExpiredPage);
		}
		if(r < 0 && errno == EINTR)
			continue;
		if(r < sz) {
			WriteLogLine ("Database problem. Corrupted");
			ShowBadPage ("Database problem. Corrupted");
		}
		memcpy(&k, content, sizeof(k));
		if(k == db_key)
			break;
	}
	memcpy(&tc, content + sizeof(k), sizeof(tc));
	if(tc == 0) {
		close(fd);
		WriteLogLine ("Code %d expired!", db_key);
		if(!DEBUG)
			return 0;
	}
	memcpy(&code, content + sizeof(k) + sizeof(tc), sizeof(code));
	if(try_key != code) {
		WriteLogLine ("try = %d code = %d", try_key, code);
		close(fd);
		if(!DEBUG)
			return 0;
	}
	tc = 0;
	lseek(fd, (off_t)-sz, SEEK_CUR);
	memcpy(content + sizeof(k), &tc, sizeof(tc));
	if(!DEBUG)
		write(fd, content, sz);
	close(fd);
	return 1;
}

#define FQDN 1
#define SUBJ 2

static char *FromToBase64(char *FromTo, int return_fqdn_subj)
{
	char *lt = NULL;
	char *s = NULL;
	char *e = NULL;
	int need_base64 = 0;
	size_t size;

	if(return_fqdn_subj == SUBJ)
		goto subj;
	lt = strchr(FromTo, '<');
	if(lt != NULL)
		s = strchr(lt + 1, '>');
	e = strchr(FromTo + 1, '"');

	if(FromTo[0] != '"' || lt == NULL || s == NULL || (lt+1) == s ||
			e == NULL || e > lt) {
err:
		WriteLogLine ("Strange \"boot name\" <email@valid.dom> "
				"(%s) configured.", FromTo);
		ShowBadPage ("Internal problem");
	}
	if(return_fqdn_subj == FQDN) {
		s = strdup(lt + 1);
		if(s == NULL) {
			WriteLogLine ("Memory exchausted.");
			ShowBadPage ("Memory exchausted.");
		}
		e = strchr(s, '>');
		*e = '\0';
		return s;
	}
  subj:
	e = NULL;
	s = FromTo + (return_fqdn_subj == 0);
	while(*s != '\0') {
		if(((int)((unsigned char)*s)) > 127) {
			if(e != NULL && return_fqdn_subj == 0)
				goto err;
			need_base64++;
			if(return_fqdn_subj == SUBJ)
				break;
		}
		if(*s == '"')
			e = s;
		s++;
	}
	if(return_fqdn_subj == 0 && e[1] != ' ')
		goto err;
	if(need_base64 == 0) {
		s = strdup(FromTo);
		if(s == NULL) {
			WriteLogLine ("Memory exchausted.");
			ShowBadPage ("Memory exchausted.");
		}
		return s;
	}
	if(return_fqdn_subj == 0)
		*++e = '\0';
	size = strlen(FromTo);
	if(return_fqdn_subj == 0)
		*e = ' ';
	else
		e = "";
	if(size > SRC_BUF_SIZE) {
		WriteLogLine ("%s too long", FromTo);
		ShowBadPage ("Internal problem");
	}
	memcpy(src_buf, FromTo, size);
	if (size < SRC_BUF_SIZE) {
		/* pad with 0s so we can just encode extra bits */
		memset (src_buf + size, 0, 3 - (size % 3));
	}
	/* Encode the buffer */
	size = b64encode (src_buf, dst_buf, size);
	dst_buf[size-1] = '\0'; /* \n */
	s = malloc(2 + strlen(charset) + 3 + size + 1 + strlen(e) + 1);
	if(s == NULL) {
		WriteLogLine ("Memory exchausted.");
		ShowBadPage ("Memory exchausted.");
	}
	sprintf(s, "=?%s?B?%s?=%s" , charset, dst_buf, e);
	return s;
}

static void SendEmail(void)
{
	FILE *Fout;
	char *fromtosubj;
	char *sendmail;
	int f;

	fromtosubj = FromToBase64(To, FQDN);
	sendmail = malloc(strlen(Sendmail) + 1 + strlen(fromtosubj) + 1);
	sprintf(sendmail, "%s %s", Sendmail, fromtosubj);
	free(fromtosubj);

	if(DEBUG) {
		while( (fromtosubj = strpbrk(sendmail, "/ <>"))  != NULL)
			*fromtosubj = '_';
		Fout = fopen(sendmail, "w");
	} else {
		Fout = popen(sendmail, "w");
	}

	if(Fout == NULL) {
		WriteLogLine ("%s: %m", Sendmail);
		ShowBadPage("Internal problem");
	}
	fromtosubj = FromToBase64(From, 0);
	fprintf(Fout, "From: %s\n", fromtosubj);
	free(fromtosubj);
	fromtosubj = FromToBase64(To, 0);
	fprintf(Fout, "To: %s\n", fromtosubj);
	free(fromtosubj);
	fprintf(Fout, "X-Mailer: %s\n", X_Mailer);
	fromtosubj = FromToBase64(Subject, SUBJ);
	fprintf(Fout, "Subject: %s\n", fromtosubj);
	free(fromtosubj);
	fprintf(Fout, "MIME-Version: 1.0\n"
"Content-Type: multipart/mixed; boundary=____boundary____\n"
"\n"
"This is a multi-part message in MIME format.\n"
"--____boundary____\n"
"Content-Type: text/plain; charset=%s\n"
"Content-Transfer-Encoding: 8bit\n\n", charset);

	for(f = Start_Form_send_to_email_body; f <= End_Form_send_to_email_body; f++) {
		if(Form[f].stored) {
			fprintf(Fout, "%s: %s\n",
					form_cfg[Form[f].cfg_num].name,
					Form[f].stored);
		}
	}
	fprintf(Fout, "--____boundary____");
	if(LastFileName[0]) {
		FILE *in;
		int c;

		fprintf(Fout, "\nContent-Type: application/octet-stream;\n"
" name=\"%s\"\n"
"Content-Transfer-Encoding: base64\n"
"Content-Disposition: attachment;\n"
" filename=\"%s\"\n\n", Attached, Attached);
		in = fopen(LastFileName, "r");
		while( (c=getc(in)) != EOF)
			putc(c, Fout);
		fclose(in);
		fprintf(Fout, "\n--____boundary____");
	}
	fprintf(Fout, "--\n");
	fclose(Fout);
}

int
main (int argc, char *argv[], char *environment[])
{
  char *p;
  long InCount;         /* The supposed number of incoming bytes. */
  long RealCount;       /* Actual number of incoming bytes. */
  char *Content;        /* Copy in memory of the Content. */
  char *Here;           /* Position into the Content. */

  if(LoadConfig(form_cfg) < 0) {
	printf("Content-type: text/plain\n\n");
	printf("LoadConfig problem\n");
	return 14;
  }
  if (DEBUG) {
	printf ("Content-type: text/html\n\n");
	printf ("<html>\n<head>\n");
	printf ("<meta http-equiv=expires content=\"0\">\n");
	printf ("</head>\n<body>\n");
  }

  /* Test if the program was started by a METHOD=POST form. */
  p = getenv ("REQUEST_METHOD");
  if ((p == NULL) || (*p == '\0') || (strcasecmp (p, "post") != 0)) {
	ShowBadPage ("Sorry, this program only supports POST.");
  }
  if (DEBUG)
	printf ("<center><h2>Loading input</h2></center>\np = %s<br>\n", p);

  /* Test if the program was started with ENCTYPE="multipart/form-data". */
  p = getenv ("CONTENT_TYPE");
  if ((p == NULL) ||
      (strncasecmp(p, "multipart/form-data; boundary=", 30) != 0)) {
	ShowBadPage("Sorry, this program only supports "
				"ENCTYPE=\"multipart/form-data\".");
  }
  if (DEBUG)
	printf ("Enctype = %s<br>\n", p);

  /* Determine the Boundary, the string that separates the segments in the
     stream. The boundary is available from the CONTENT_TYPE environment
     variable. */
  Here = strchr (p, '=') - 1;
  strlen_Boundary = strlen(Here);
  Boundary = Here;
  Boundary[0] = '-'; Boundary[1] = '-';
  if (DEBUG)
	printf ("Boundary = %s<br>\n", Boundary);

  /* Get the total number of bytes in the input stream from the
     CONTENT_LENGTH environment variable. */
  p = getenv ("CONTENT_LENGTH");
  if (p == NULL) {
	WriteLogLine ("Error: no CONTENT_LENGTH found.");
	ShowBadPage ("Error: no CONTENT_LENGTH found.");
  }
  InCount = atol (p);
  if (InCount == 0) {
	WriteLogLine ("Error: CONTENT_LENGTH is zero.");
	ShowBadPage ("Error: CONTENT_LENGTH is zero.");
  }
  if (DEBUG)
	printf ("Content_Length = %ld<br>\n", InCount);

  /* Allocate sufficient memory for the incoming data. */
  Content = (char *) malloc (InCount + 1);
  if (Content == NULL) {
	WriteLogLine ("Error: malloc() returned NULL.");
	ShowBadPage ("Error: malloc() returned NULL.");
  }

  /* Load the data from standard input into memory. */
  RealCount = fread (Content, 1, InCount, stdin);
  if (RealCount != InCount) {
	WriteLogLine ("Error: The number of bytes received (%ld) "
			"is not what the CONTENT_LENGTH environment variable "
			"says it sould be (%ld).",
			 RealCount, InCount);
	ShowBadPage ("Error: The number of bytes received (%ld) "
			"is not what the CONTENT_LENGTH environment variable "
			"says it sould be (%ld).",
			 RealCount, InCount);
	return 1;
  }
  /* Ignore any extra caracters. We have to read them, or the server
     will give an error message. */
  while(getchar() != EOF)
	;
  if (DEBUG) {
	printf ("Input succesfully loaded into memory.<br>\n");
	if (DEBUG > 1) {
		FILE *out = fopen("/tmp/mail_mem_out.txt", "w");
		if(out) {
			fwrite(Content, 1, InCount, out);
			fclose(out);
		}
	}
  }

  /* Handle all segments in the incoming data, that has been stored in
     memory. */
  Here = Content;
  if (DEBUG) {
	printf ("<p>\n<center>\n<h2>Parsing input</h2>\n");
	printf ("<table border=1>\n");
	printf ("<tr>\n");
	printf ("  <th>Fieldname</th>\n");
	printf ("  <th>Contents</th>\n");
	printf ("  <th>Size</th>\n");
	printf ("  </tr>\n");
  }
  do {
	Here = AcceptSegment (Here, &RealCount);
  } while (Here);
  if (DEBUG)
	printf ("</table>\n</center>\n");

  DecodeNonEmpty();
  if(get_db(atoi(Form[Form_num].stored), atoi(Form[Form_try].stored))) {
	SendEmail();
	/* Display the OkPage. */
	if (DEBUG)
		printf ("<p><center><h2>Finished</h2></center>\n");
	ShowOkPage (OkPage);
  } else {
	ShowOkPage (ExpiredPage);
  }
  return 0;
}
